# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

import logging
import minty
import re
import statsd
from minty import Base
from minty.config.parser.apache import ApacheConfigParser
from minty.infrastructure import _parse_global_config
from minty.logging.mdc import get_mdc, mdc
from minty_pyramid.renderer import jsonapi
from pyramid.config import Configurator
from uuid import uuid4


def new_response_callback(event):
    """Set default headers when a new response is created"""
    if not event.response.headers.get("cache-control"):
        event.response.headers.update({"cache-control": "no-store"})

    event.response.headers.update(
        {"request-id": str(event.request.request_id)}
    )


class GenericEngine(Base):
    """Pyramid configurator for Non Domain Driven Projects."""

    def __init__(self):

        self.config = None

    def setup(self, global_config: dict, **settings) -> object:
        """Set up the application by loading the injecting the CQRS layer.

        :param global_config: Global configuration
        :type global_config: dict
        :return: Returns the Configurator from Pyramid
        :rtype: object
        """

        config = Configurator(settings=settings)

        config.add_request_method(
            lambda r: uuid4(), name="request_id", property=True, reify=True
        )

        configuration_parsed = _parse_global_config(
            settings["minty_service.infrastructure.config_file"],
            ApacheConfigParser(),
        )
        config.add_request_method(
            lambda r: configuration_parsed,
            name="configuration",
            property=True,
            reify=True,
        )

        ### Add renderer
        config.add_renderer(None, jsonapi())

        config.add_tween("minty_pyramid.generic_loader.RequestLoadGauge")
        config.add_tween("minty_pyramid.generic_loader.RequestTimer")
        config.add_tween("minty_pyramid.generic_loader.RequestDataLogger")

        config.add_subscriber(
            "minty_pyramid.generic_loader.new_response_callback",
            "pyramid.events.NewResponse",
        )

        config.scan()
        self.config = config

        if "statsd" in configuration_parsed:
            if "disabled" in configuration_parsed["statsd"]:
                configuration_parsed["statsd"]["disabled"] = bool(
                    int(configuration_parsed["statsd"]["disabled"])
                )
            statsd.Connection.set_defaults(**configuration_parsed["statsd"])
        else:
            # No statsd configuration available; forcefully disable it
            statsd.Connection.set_defaults(disabled=True)

        return config

    def main(self) -> object:
        """Run the application by calling the wsgi_app function of Pyramid.

        :raises ValueError: When setup is forgotten
        :return: wsgi app
        :rtype: object
        """
        if self.config is None:
            raise ValueError("Make sure you run setup before 'main'")

        self.logger.info("Creating WSGI application")

        return self.config.make_wsgi_app()


class RequestLoadGauge(Base):
    """
    Pyramid "tween" class to report load (number of active threads).

    This uses a statsd gauge to report the number of concurrently active
    threads in this process to the platform.
    """

    def __init__(self, handler, registry):
        self.handler = handler
        self.registry = registry

    def __call__(self, request):
        # We want a "context free" statsd, so we do not split out the
        # metrics by hostname.

        old_context = self.statsd.context
        self.statsd.context = None

        self.statsd.get_counter("total_requests").increment()
        gauge = self.statsd.get_gauge("active_threads")

        self.statsd.context = old_context

        try:
            gauge.increment()
            response = self.handler(request)
        finally:
            gauge.decrement()

        return response


class RequestTimer(Base):
    """Middleware / tween to log request time and count to statsd."""

    def __init__(self, handler, registry):
        self.handler = handler
        self.registry = registry

    def __call__(self, request):
        "Log time and count to statsd before and after execution of handler."

        try:
            # Pre-build the timer & counter objects, so it doesn't matter if
            # `self.handler` (re)sets the context in `minty.STATSD`.
            minty.STATSD.context = request.host.replace(".", "_")
            timer = self.statsd.get_timer("request_duration")
            counter = self.statsd.get_counter("number_of_requests")

            metrics_path = re.sub(r"[^a-zA-Z0-9_]", "_", request.path[1:])

            with timer.time(metrics_path):
                response = self.handler(request)

            counter.increment(str(response.status_int), delta=1)
        finally:
            minty.STATSD.context = None

        return response


def get_logging_info(request):
    logging_info = {}

    logging_info["hostname"] = request.host
    logging_info["request_id"] = request.request_id
    logging_info["action_path"] = request.path_info

    # Log the instance_hostname for this configuration, if available;
    # otherwise the hostname used by the user is used.
    logging_info["instance_hostname"] = request.configuration.get(
        "instance_hostname", request.host
    )

    try:
        logging_info["session_id"] = request.session_id
    except AttributeError:
        logging_info["session_id"] = "<sessions not configured>"

    return logging_info


class RequestDataLogger(Base):
    """Tween to add request-specific information to log entries"""

    def __init__(self, handler, registry):
        self.handler = handler
        self.registry = registry

    def __call__(self, request):
        logging_info = get_logging_info(request)
        with mdc(**logging_info):
            response = self.handler(request)

        return response


# Add a "req" key, containing the MDC to logging output in minty-pyramid
# applications.
old_factory = logging.getLogRecordFactory()


def log_record_factory(*args, **kwargs):
    record = old_factory(*args, **kwargs)
    record.req = get_mdc()
    return record


logging.setLogRecordFactory(log_record_factory)



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

import logging
from .exception_handlers import (
    exception_cqrs_catch_all,
    exception_cqrs_conflict,
    exception_cqrs_forbidden,
    exception_cqrs_not_found,
    exception_cqrs_validation,
    exception_cqrs_validation_pydantic,
)
from .generic_loader import GenericEngine, log_record_factory
from .session_manager import SessionRetriever, session_manager_factory
from minty.cqrs import CQRS
from minty.infrastructure import InfrastructureFactory
from minty_pyramid.renderer import jsonapi
from pyramid.config import Configurator
from typing import Optional, Union
from uuid import UUID, uuid4


class Engine(GenericEngine):
    """Pyramid configurator."""

    def __init__(
        self,
        domains: list,
        command_wrapper_middleware: Optional[list] = None,
        query_middleware: Optional[list] = None,
    ):
        if command_wrapper_middleware is None:
            command_wrapper_middleware = []
        if query_middleware is None:
            query_middleware = []
        self.command_wrapper_middleware = command_wrapper_middleware
        self.query_middleware = query_middleware
        self.domains = domains
        self.config = None

    def setup(self, global_config: dict, **settings) -> object:
        """Set up the application by loading the injecting the CQRS layer.

        :param global_config: Global configuration
        :type global_config: dict
        :return: Returns the Configurator from Pyramid
        :rtype: object
        """
        infra_factory = InfrastructureFactory(
            settings["minty_service.infrastructure.config_file"]
        )
        config = Configurator(settings=settings)

        cqrs = CQRS(
            domains=self.domains,
            infrastructure_factory=infra_factory,
            command_wrapper_middleware=self.command_wrapper_middleware,
            query_middleware=self.query_middleware,
        )

        config.add_request_method(
            lambda r: uuid4(), name="request_id", property=True, reify=True
        )

        config.add_request_method(
            lambda r, n: cqrs.infrastructure_factory.get_infrastructure(
                context=r.host, infrastructure_name=n
            ),
            name="get_infrastructure",
            property=False,
        )

        config.add_request_method(
            lambda r, n: cqrs.infrastructure_factory_ro.get_infrastructure(
                context=r.host, infrastructure_name=n
            ),
            name="get_infrastructure_ro",
            property=False,
        )

        config.add_request_method(
            lambda r: infra_factory,
            name="infrastructure_factory",
            property=True,
            reify=True,
        )

        config.add_request_method(
            lambda r: r.infrastructure_factory.get_config(context=r.host),
            name="configuration",
            property=True,
            reify=True,
        )

        ### Add renderer
        config.add_renderer(None, jsonapi())

        config.include(_build_cqrs_setup(cqrs))
        if settings.get("session_manager", False):
            session_manager = session_manager_factory(
                infra_factory=infra_factory
            )
            config.include(
                _build_http_session_manager(
                    session_manager=session_manager,
                    cookie_name=settings.get(
                        "session_cookie_name", "minty_session"
                    ),
                )
            )

        config.add_view(view=exception_cqrs_not_found, name="not_found")
        config.add_view(view=exception_cqrs_forbidden, name="forbidden")
        config.add_view(view=exception_cqrs_conflict, name="conflict")
        config.add_view(view=exception_cqrs_validation, name="validation")
        config.add_view(
            view=exception_cqrs_validation_pydantic, name="validation_pydantic"
        )
        config.add_view(view=exception_cqrs_catch_all, name="catch_all")

        config.add_tween("minty_pyramid.generic_loader.RequestLoadGauge")
        config.add_tween("minty_pyramid.generic_loader.RequestTimer")
        config.add_tween("minty_pyramid.generic_loader.RequestDataLogger")

        config.add_subscriber(
            "minty_pyramid.generic_loader.new_response_callback",
            "pyramid.events.NewResponse",
        )

        config.scan()
        self.config = config
        return config


def _build_cqrs_setup(cqrs: CQRS):
    """Create a callable for setting up the "CQRS" methods request objects.

    :param cqrs: A configured CQRS object
    :type cqrs: CQRS
    :return: A function, callable by Pyramid, to register the CQRS
        method(s)
    :rtype: callable
    """

    def setup_cqrs_request(config):
        """Add the CQRS accessors to the Pyramid request objects.

        :param config: Pyramid configurator instance
        :type config: Configurator
        :return: Nothing
        :rtype: None
        """

        def get_query_instance(
            request, domain: str, user_uuid: Union[UUID, str], user_info=None
        ):
            validated_user_uuid = UUID(str(user_uuid))

            cqrs_instance = cqrs.get_query_instance(
                correlation_id=request.request_id,
                domain=domain,
                context=request.host,
                user_uuid=validated_user_uuid,
                user_info=user_info,
            )
            return cqrs_instance

        config.add_request_method(get_query_instance, "get_query_instance")

        def get_command_instance(
            request, domain: str, user_uuid: Union[UUID, str], user_info=None
        ):
            validated_user_uuid = UUID(str(user_uuid))

            cqrs_instance = cqrs.get_command_instance(
                correlation_id=request.request_id,
                domain=domain,
                context=request.host,
                user_uuid=validated_user_uuid,
                user_info=user_info,
            )
            return cqrs_instance

        config.add_request_method(get_command_instance, "get_command_instance")

    return setup_cqrs_request


def _build_http_session_manager(
    session_manager: SessionRetriever,
    cookie_name: str = "minty_cookie",
):
    """Create a callable to set up the `retrieve_session` method on request objects.

    :param session_manager: A configured SessionRetriever object
    :type session_manager: SessionRetriever
    :param cookie_name: Name of the cookie the session-id is stored in
    :type cookie_name: str
    :return: A function, callable by Pyramid, to register the session_manager
        method(s)
    :rtype: callable
    """

    def setup_session_request(config):
        """Add the SessionRetriver accessors to the Pyramid request objects.

        :param config: Pyramid configurator instance
        :type config: Configurator
        :return: Nothing
        :rtype: None
        """

        def session_id(request):
            try:
                return request.cookies[cookie_name]
            except KeyError:
                return "<session cookie not found>"

        config.add_request_method(
            session_id, "session_id", property=True, reify=True
        )

        def retrieve_session(request):
            session = session_manager.retrieve(request.cookies[cookie_name])
            return session

        config.add_request_method(retrieve_session, "retrieve_session")

    return setup_session_request


# Add a "req" key, containing the MDC to logging output in minty-pyramid
# applications.
old_factory = logging.getLogRecordFactory()
logging.setLogRecordFactory(log_record_factory)



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

import logging
import re
from minty.cqrs import UserInfo
from minty.entity import RedirectResponse
from minty.validation import assert_never
from minty_pyramid.session_manager import (
    check_user_permission,
    get_logged_in_user,
)
from pyramid.httpexceptions import HTTPBadRequest, HTTPForbidden
from pyramid.response import Response
from typing import Dict, Literal, Optional, Set, TypedDict, Union
from uuid import UUID, uuid4

log = logging.getLogger(__name__)

# Default page size if none is specified.
# Note that this is a str() because it's a default for a query parameter.
DEFAULT_PAGE_SIZE = str(50)


SupportedHTTPMethods = Literal["GET", "POST"]
CQType = Literal["command", "query"]


# Until PEP-655 lands, we need this workaround for optional keys in a TypedDict
class FromDefinition(TypedDict, total=False):
    request_params: dict
    json: dict


class PagingDefinitionBase(TypedDict):
    max_results: int


class PagingDefinition(PagingDefinitionBase, total=False):
    page: Literal["page"]
    page_size: Literal["page_size"]


ViewMappingBase = TypedDict(
    "ViewMappingBase",
    {
        "cq": CQType,
        "auth_permissions": Set[str],
        "domain": str,
        "run": str,
        "from": FromDefinition,
    },
)


class ViewMapping(ViewMappingBase, total=False):
    entity_type: str
    command_primary: str
    paging: PagingDefinition


ViewMapper = Dict[SupportedHTTPMethods, Dict[str, ViewMapping]]


# This will be moved to minty-pyramid in a next iteration
class JSONAPIEntityView:
    """
    View Base class for Pyramid Views which return Entity objects

    Use this as a base class for your own view. E.g.

        class DogView(JSONAPIEntityView):
            view_mapper = {
                "GET": {
                    "get_dog": {
                        "cq": "query",
                        "auth_permissions": ["gebruiker"],
                        "domain": "zsnl_domains.case_management",
                        "run": "get_dog_by_uuid",
                        "from": {"request_params": {"uuid": "uuid"}},
                    },
                    "filter_dogs": {
                        "cq": "query",
                        "auth_permissions": ["gebruiker"],
                        "domain": "zsnl_domains.case_management",
                        "run": "get_dog_by_uuid",
                        "from": {"request_params": {"filter[]": "filter_params"}},
                    }
                },
                "POST": {
                    "create_dog": {
                        "cq": "command",
                        "auth_permissions": ["gebruiker"],
                        "domain": "zsnl_domains.case_management",
                        "run": "create_dog",
                        "from": {
                            "request_params": {"uuid": "uuid"},
                            "json": {"_all": True},
                        },
                        "command_primary": "uuid",
                        "entity_type": "dog",
                    }
                }
            }

            def create_link_from_entity(
                self, entity=None, entity_type=None, entity_id=None
            ):
                if entity is not None:
                    entity_type = entity.entity_type
                    entity_id = str(entity.entity_id)

                if entity_type == "dog":
                    return self.request.route_url(
                        "get_dog_object", _query={"uuid": entity_id}, _scheme="https"
                    )

    The view_mappers has as it's first key the method of the API call, e.g.
    POST or GET. Define the action in the url as the first key of the following
    dict (get_dog, create_dog) in the above example.

    Other parameters:
        cq (Enum("query", "command")):  Defines whether to call on the Query or
                                        Command class
        auth_permissions (Set[str]): The list of permissions to check
        domain (str): The domain to run this command or query against
        run (str): the command or query to run
        from (dict): A dict of parameters, see below
        command_primary (str):  The "reference" of the command extracted from
                                the given "parameter
        entity_type (str):  The string entity_type, so we know when we got no
                            Entity as a return value, which type to set
        generate_missing_id(str| list(str)): Argument(s) for which a id (uuid)
                            will be generated if not in the request_params.

    The "from" parameter extracts input parameters from either the given
    "request_parameters" or the "json_body" of the request. The following
    dict contains a key value pair. The key correspondents to the key from the
    input. The value correspondents to the key for the command to run.

    Using array[] options in request_params:
    When a key of a request_params ends with the characters [] (e.g.: see
    filter_dogs above), then it will transform the variables into a dict.

    For instance: ?filter[name]=frits&filter[gender]=male
    Will transform into:
        {
            "name": "frits",
            "gender": "male"
        }

    """

    view_mapper: ViewMapper

    def __init__(self, context, request):
        """Initializes this object from Pyramid"""

        self.request = request
        self.call = request.matched_route.name

    def _check_protected_route(self, calldata):
        """Verifies authentication permissions against our session"""

        user_info = get_logged_in_user(self.request, log)
        if not check_user_permission(calldata["auth_permissions"], user_info):
            raise HTTPForbidden(
                json={"error": "You do not have permission to access this."}
            )

        return user_info

    def _get_cqrs_instance(self, calldata: ViewMapping, user_info: UserInfo):
        """Gets the command or query instance for the called api view"""
        cq = calldata["cq"]

        cmdqry = getattr(self.request, f"get_{cq}_instance")
        return cmdqry(
            domain=calldata["domain"],
            user_info=user_info,
            user_uuid=user_info.user_uuid,
        )

    def __call__(self):
        """Reads and runs the called API route defined in view_mapper"""

        calldata = self.view_mapper[self.request.method][self.call]

        user_info = self._check_protected_route(calldata)
        cqinstance = self._get_cqrs_instance(calldata, user_info)

        run = getattr(cqinstance, calldata["run"])
        params = self._collect_params(calldata)

        response = run(**params)

        return self.transform_response_from_command_or_query(
            response, calldata, params
        )

    def transform_response_from_entity(
        self, entity_type: str, entity_id: UUID
    ):
        """Transform response from entity"""

        link = self.create_link_from_entity(
            entity_type=entity_type,
            entity_id=entity_id,
        )

        if link:
            links_block = {"links": {"self": link}}
        else:
            links_block = {}

        return {
            **links_block,
            "type": entity_type,
            "id": entity_id,
        }

    def transform_response_from_command(
        self, calldata: ViewMapping, params
    ) -> dict:
        """Transform response from command"""

        if "command_primary" not in calldata or "entity_type" not in calldata:
            return {"data": None}

        if params.get(calldata["command_primary"], None):
            call_definition_entity_type = calldata["entity_type"]

            try:
                entity_type = params[call_definition_entity_type]
            except KeyError:
                entity_type = call_definition_entity_type

            if isinstance(params[calldata["command_primary"]], list):
                res = []
                for entity_id in params[calldata["command_primary"]]:
                    res.append(
                        self.transform_response_from_entity(
                            entity_type, entity_id
                        )
                    )
            else:
                entity_id = params[calldata["command_primary"]]
                res = self.transform_response_from_entity(
                    entity_type, entity_id
                )
            return {"data": res}
        else:
            return {"data": None}

    def transform_response_from_query(self, response) -> Union[dict, Response]:
        """Transform response from query"""
        if response is None:
            return {"data": None}
        elif isinstance(response, RedirectResponse):
            return Response(status_int=302, location=response.location)
        else:
            json_response = {
                "data": response,
                "links": {"self": self.request.current_route_path()},
            }
            self._apply_included_entities(response, json_response)
            self._apply_meta(response, json_response)
            return json_response

    def transform_response_from_command_or_query(
        self, response, calldata: ViewMapping, params
    ) -> Optional[Union[dict, Response]]:
        """
        Transforms a response from a command to a proper JSONApi structure

        Args:
            response (Optional[Entity]): The response from our query or command
            calldata (dict): The configuration defined in our view_mapper
            params (dict): The collected parameters from our query and jsonbody

        Returns:
            Structured dict for our jsonapi serializer:

                {
                    "data": {
                        "links": {
                            "self": "http://url",
                        },
                        "type": "name_of_entity",
                        "id": "12434-abcde-1234-acbewe-233aba"
                    }
                }

        """
        if calldata["cq"] == "command":
            return self.transform_response_from_command(calldata, params)
        elif calldata["cq"] == "query":
            return self.transform_response_from_query(response)

        assert_never(calldata["cq"])  # pragma: no cover

    def _apply_meta(self, response, json_response: dict):
        """Add meta data to json_response if available"""

        if (
            hasattr(response, "total_results")
            and response.total_results is not None
        ):
            json_response["meta"] = {"total_results": response.total_results}

    def _apply_included_entities(self, response, json_response: dict):
        if (
            hasattr(response, "included_entities")
            and response.included_entities is not None
        ):
            json_response["included"] = response.included_entities

    def _collect_params_as_array(self, k, v, params):
        for pk, pv in self.request.params.items():
            match = re.search(r"^" + k + r"\[([^\]]+)\]$", pk)

            if match:
                if v not in params:
                    params[v] = {}

                params[v][match[1]] = pv

    def _convert_value(self, request_param, domain_param_config):
        param_key = domain_param_config["name"]
        param_type = domain_param_config["type"]

        try:
            param_value = param_type(request_param)
        except ValueError:
            # View mapper defined a type, but the parameter can't be converted from
            # string to that type.
            raise HTTPBadRequest

        return (param_key, param_value)

    def _extract_request_params(self, calldata, request_params):
        final_params = {}

        for http_param_name, domain_param_name in calldata["from"][
            "request_params"
        ].items():
            # Special array functionality
            if http_param_name.endswith("[]"):
                http_param_name = http_param_name.replace("[]", "")
                self._collect_params_as_array(
                    http_param_name, domain_param_name, final_params
                )

                continue

            if http_param_name not in request_params:
                continue

            if isinstance(domain_param_name, dict):
                (domain_param, converted_value) = self._convert_value(
                    request_param=request_params[http_param_name],
                    domain_param_config=domain_param_name,
                )
                final_params[domain_param] = converted_value
            else:
                final_params[domain_param_name] = request_params[
                    http_param_name
                ]

        return final_params

    def __collect_missing_ids(self, calldata, params: dict):
        if "generate_missing_id" in calldata:
            if isinstance(calldata["generate_missing_id"], list):
                for id in calldata["generate_missing_id"]:
                    if id not in params:
                        params[id] = uuid4()
            elif calldata["generate_missing_id"] not in params:
                params[calldata["generate_missing_id"]] = uuid4()

    def _collect_params(self, calldata):
        """Collects parameters according to the configuration in view_mapper"""
        params = {}
        if "request_params" in calldata["from"]:
            params = self._extract_request_params(
                calldata, self.request.params
            )

        if "json" in calldata["from"]:
            if (
                "_all" in calldata["from"]["json"]
                and calldata["from"]["json"]["_all"] is True
            ):
                for k in self.request.json_body.keys():
                    params[k] = self.request.json_body[k]
            else:
                for k, v in calldata["from"]["json"].items():
                    if k not in self.request.json_body:
                        continue
                    params[v] = self.request.json_body[k]

        if "paging" in calldata:
            paging_params = self._extract_paging_params(
                calldata, self.request.params
            )
            params.update(paging_params)

        self.__collect_missing_ids(calldata, params)

        return params

    def create_link_from_entity(
        self, entity=None, entity_type=None, entity_id=None
    ) -> Optional[str]:
        log.debug(
            "create_link_from_entity called, but not overridden in view class"
        )
        return None

    def _extract_paging_params(self, calldata, request_params):
        final_params = {}
        final_params["page"] = request_params.get("page", "1")
        final_params["page_size"] = request_params.get(
            "page_size", DEFAULT_PAGE_SIZE
        )

        self._check_paging_limits(
            int(final_params["page"]),
            int(final_params["page_size"]),
            int(calldata["paging"]["max_results"]),
        )
        return final_params

    def _check_paging_limits(
        self, page: int, page_size: int, max_results: int
    ):
        if page < 1:
            raise HTTPBadRequest(
                json={"errors": [{"title": "'page' must be >= 1"}]}
            )

        if page_size < 1:
            raise HTTPBadRequest(
                json={"errors": [{"title": "'page_size' must be >= 1"}]}
            )

        if page_size > max_results:
            raise HTTPBadRequest(
                json={
                    "errors": [
                        {
                            "title": f"Requested page size {page_size} > max size {max_results}."
                        }
                    ]
                }
            )



# SPDX-FileCopyrightText: 2020 Mintlab B.V.
#
# SPDX-License-Identifier: EUPL-1.2

import datetime
import logging
from minty.entity import Entity, EntityCollection, ValueObject
from pyramid.renderers import JSON
from typing import Any, Dict
from uuid import UUID

logger = logging.getLogger(__name__)


class jsonapi(JSON):
    jsonapi_adapters = {
        Entity: "_entity_adapter",
        EntityCollection: "_entity_collection_adapter",
        ValueObject: "_value_object_adapter",
        set: "_set_adapter",
        UUID: "_uuid_adapter",
        datetime.datetime: "_datetime_adapter",
        datetime.date: "_datetime_adapter",
    }
    view: Any

    def __init__(self, *args, **kwargs):
        """Constructs the jsonapi class based on Pyramid JSON implementation.

        This will add some default adapters to the system, to handle entities,
        value objects, etc.
        """

        super().__init__(*args, **kwargs)

        for obj, adapter in self.jsonapi_adapters.items():
            self.add_adapter(obj, getattr(self, adapter))

    def __call__(self, info):
        """Returns a plain JSON-encoded string with content-type
        ``application/json``. The content-type may be overridden by
        setting ``request.response.content_type``."""

        # We had top copy this from the Pyramid JSON implementation, because
        # we would like to inject the view on self, so we can use the view
        # class in adapters for generating urls for example
        def _render(value, system):
            request = system.get("request")
            self.view = system.get("view")

            if request is not None:
                response = request.response
                ct = response.content_type
                if ct == response.default_content_type:
                    response.content_type = "application/json"
            default = self._make_default(request)

            return self.serializer(value, default=default, **self.kw)

        return _render

    def _entity_collection_adapter(self, obj, request):
        return [entity for entity in obj]

    def __links_for_entity(self, entity):
        if not hasattr(self.view, "create_link_from_entity"):
            logger.warning(f"Unknown kind of view: {self.view.__module__}")
        self_link = self.view.create_link_from_entity(entity)
        if self_link:
            return {"links": {"self": self_link}}
        else:
            return {}

    def __add_nested_relationships(
        self, relationshipcontent, relationship_response
    ):
        # Retrieve nested relationships within a relationship
        if getattr(relationshipcontent, "entity_relationships", []):
            # Add new field 'relationships' to relationship_response for nested relationships
            relationship_response["relationships"] = {}

            for (
                nested_relationship_name
            ) in relationshipcontent.entity_relationships:
                nested_relationship = getattr(
                    relationshipcontent, nested_relationship_name
                )

                if nested_relationship is None:
                    continue

                # Fill in meta data for nested relationship
                metadata = {}
                if nested_relationship.entity_meta_summary is not None:
                    metadata["meta"] = {
                        "summary": nested_relationship.entity_meta_summary
                    }

                # Add nested relationship to 'relationships' field of relationship_response
                relationship_response["relationships"].update(
                    {
                        nested_relationship_name: {
                            "data": {
                                "type": nested_relationship.entity_type,
                                "id": nested_relationship.entity_id,
                            },
                            **metadata,
                            **self.__links_for_entity(nested_relationship),
                        }
                    }
                )

    def __build_relationships_for_entity(self, obj: Entity, object_attributes):
        object_relationships: Dict[str, Any] = {}

        # Retrieve relationships from entity
        for relationshipkey in obj.entity_relationships:
            relationship = getattr(obj, relationshipkey)
            if relationship is not None:
                list_of_relationships = []
                is_list = None
                if isinstance(relationship, list):
                    list_of_relationships = relationship
                    is_list = True
                else:
                    list_of_relationships = [relationship]

                for relationshipcontent in list_of_relationships:
                    relationshipdata = relationshipcontent.dict()
                    metadata = {}
                    meta_content = self._build_meta(relationshipcontent)
                    if meta_content:
                        metadata["meta"] = meta_content

                    relationship_response = {
                        "data": {
                            "type": relationshipdata["entity_type"],
                            "id": relationshipdata["entity_id"],
                        },
                        **metadata,
                        **self.__links_for_entity(relationshipcontent),
                    }

                    self.__add_nested_relationships(
                        relationshipcontent, relationship_response
                    )

                    if not is_list:
                        object_relationships[
                            relationshipkey
                        ] = relationship_response
                    else:
                        if relationshipkey not in object_relationships:
                            object_relationships[relationshipkey] = []

                        object_relationships[relationshipkey].append(
                            relationship_response
                        )

            del object_attributes[relationshipkey]
        return object_relationships

    def _entity_adapter(self, obj: Entity, request):
        """Transforms an entity in a JSON:API styled format

        See renderers documentation of Pyramid for more information about
        using adapters
        """

        object_attributes = obj.entity_dict()
        object_relationships = self.__build_relationships_for_entity(
            obj, object_attributes
        )

        for entity_id__field in obj.entity_id__fields:
            del object_attributes[entity_id__field]

        relationships_block = {}
        if object_relationships != {}:
            relationships_block["relationships"] = object_relationships

        result = {
            "type": obj.entity_type,
            "id": obj.entity_id,
            "attributes": object_attributes,
            **relationships_block,
            **self.__links_for_entity(obj),
        }
        meta = self._build_meta(obj)
        if meta:
            result["meta"] = meta

        return result

    def _build_meta(self, obj: Entity) -> dict:
        meta = {}
        for field_name in obj.entity_meta__fields:
            short_field_name = field_name.replace("entity_meta_", "")
            value = getattr(obj, field_name)
            if value is not None:
                meta[short_field_name] = value

        return meta

    def _set_adapter(self, obj, request):
        return list(sorted(obj))

    def _uuid_adapter(self, obj, request):
        """Transforms a UUID in a JSONApi styled format"""
        return str(obj)

    def _datetime_adapter(self, obj, request):
        """Transforms a datetime object in a JSONApi styled format"""
        return obj.isoformat()

    def _value_object_adapter(self, obj: ValueObject, request):
        """Transforms a ValueObject in a JSONApi styled format

        See renderers documentation of Pyramid for more information about
        using adapters
        """
        logger = logging.getLogger(__name__)
        logger.info(f"XXX {obj}")
        return obj.dict(by_alias=True)
